#-*-coding:utf-8;-*-
from errno import EISDIR,ENOENT
from json import dumps,load
from os.path import abspath,dirname,exists,expandvars,isfile,join
from socket import AF_INET,SOCK_STREAM,SocketType,socket
from subprocess import run
from tempfile import NamedTemporaryFile
from time import time_ns
from typing import Any,Optional
from urllib.parse import quote,urlunsplit
MODULE_PATH=dirname(__spec__.origin)
CONFIG=load(open(join(MODULE_PATH,"config.json"),"r",encoding="utf-8"))
AUTO_RUNNER=open(join(MODULE_PATH,"autorunner.js"),"r",encoding="utf-8").read()
FILE_RUNNER=open(join(MODULE_PATH,"filerunner.js"),"r",encoding="utf-8").read()
STRING_RUNNER=open(join(MODULE_PATH,"stringrunner.js"),"r",encoding="utf-8").read()
def bindAvailablePort(unboundSocket:SocketType,listenBacklog:Optional[int]=None,connectAddress:Any=None)->int:
    port=CONFIG["min_port"]
    while True:
        address=(CONFIG["host_name"],port)
        try:
            unboundSocket.bind(address)
        except OSError:
            if port<CONFIG["max_port"]:
                port+=1
            else:
                raise OverflowError("No available ports found")
        else:
            break
    if connectAddress is None:
        if listenBacklog is not None:
            unboundSocket.listen(int(listenBacklog))
    else:
        unboundSocket.connect(connectAddress)
    return port
def runAutoFile(filePath:str)->None:
    absolutePath=abspath(str(filePath))
    if not exists(absolutePath):
        raise FileNotFoundError(ENOENT,"The file doesn't exist",filePath)
    if not isfile(absolutePath):
        raise IsADirectoryError(EISDIR,"The path belongs to a directory",filePath)
    with socket(AF_INET,SOCK_STREAM) as serverSocket:
        serverPort=bindAvailablePort(serverSocket,1)
        with NamedTemporaryFile("w",encoding="utf-8",suffix=CONFIG["temporary_file_suffix"],dir=abspath(expandvars(CONFIG["temporary_path"]))) as tempFile:
            tempFile.write(AUTO_RUNNER%(CONFIG["script_host_name"],serverPort))
            tempFile.flush()
            run((expandvars(CONFIG["am_command"]),CONFIG["am_subcommand"],"--user",CONFIG["am_user"],"-a",CONFIG["intent_action"],"-d",urlunsplit((CONFIG["url_scheme"],"",quote(tempFile.name,encoding="utf-8"),"","")),"-t",CONFIG["intent_mime_type"],"--grant-read-uri-permission","--grant-write-uri-permission","--grant-prefix-uri-permission","--include-stopped-packages","--activity-exclude-from-recents","--activity-no-animation",CONFIG["intent_component"]),check=True)
            with serverSocket.accept()[0] as clientSocket:
                clientSocket.sendall((dumps({"file":absolutePath,"path":dirname(absolutePath)},ensure_ascii=False,separators=(",",":"))+"\n").encode("utf-8"))
def runFile(filePath:str)->None:
    absolutePath=abspath(str(filePath))
    if not exists(absolutePath):
        raise FileNotFoundError(ENOENT,"The file doesn't exist",filePath)
    if not isfile(absolutePath):
        raise IsADirectoryError(EISDIR,"The path belongs to a directory",filePath)
    with socket(AF_INET,SOCK_STREAM) as serverSocket:
        serverPort=bindAvailablePort(serverSocket,1)
        with NamedTemporaryFile("w",encoding="utf-8",suffix=CONFIG["temporary_file_suffix"],dir=abspath(expandvars(CONFIG["temporary_path"]))) as tempFile:
            tempFile.write(FILE_RUNNER%(CONFIG["script_host_name"],serverPort))
            tempFile.flush()
            run((expandvars(CONFIG["am_command"]),CONFIG["am_subcommand"],"--user",CONFIG["am_user"],"-a",CONFIG["intent_action"],"-d",urlunsplit((CONFIG["url_scheme"],"",quote(tempFile.name,encoding="utf-8"),"","")),"-t",CONFIG["intent_mime_type"],"--grant-read-uri-permission","--grant-write-uri-permission","--grant-prefix-uri-permission","--include-stopped-packages","--activity-exclude-from-recents","--activity-no-animation",CONFIG["intent_component"]),check=True)
            with serverSocket.accept()[0] as clientSocket:
                clientSocket.sendall((dumps({"file":absolutePath,"path":dirname(absolutePath)},ensure_ascii=False,separators=(",",":"))+"\n").encode("utf-8"))
def runString(commandString:str,commandTitle:Optional[str]=None)->None:
    with socket(AF_INET,SOCK_STREAM) as serverSocket:
        serverPort=bindAvailablePort(serverSocket,1)
        with NamedTemporaryFile("w",encoding="utf-8",suffix=CONFIG["temporary_file_suffix"],dir=abspath(expandvars(CONFIG["temporary_path"]))) as tempFile:
            tempFile.write(STRING_RUNNER%(CONFIG["script_host_name"],serverPort))
            tempFile.flush()
            run((expandvars(CONFIG["am_command"]),CONFIG["am_subcommand"],"--user",CONFIG["am_user"],"-a",CONFIG["intent_action"],"-d",urlunsplit((CONFIG["url_scheme"],"",quote(tempFile.name,encoding="utf-8"),"","")),"-t",CONFIG["intent_mime_type"],"--grant-read-uri-permission","--grant-write-uri-permission","--grant-prefix-uri-permission","--include-stopped-packages","--activity-exclude-from-recents","--activity-no-animation",CONFIG["intent_component"]),check=True)
            with serverSocket.accept()[0] as clientSocket:
                if commandTitle is None:
                    usedTitle="%s-%d"%(CONFIG["command_title"],time_ns())
                else:
                    usedTitle=str(commandTitle)
                clientSocket.sendall((dumps({"name":usedTitle,"script":str(commandString)},ensure_ascii=False,separators=(",",":"))+"\n").encode("utf-8"))


#-*-coding:utf-8;-*-
from json import dumps,loads
from os.path import dirname,join
from socket import AF_INET,SOCK_STREAM,SocketType,socket
from time import time_ns
from typing import Any,Optional
from .runner import CONFIG,bindAvailablePort,runString
REMOTE_CALLER=open(join(dirname(__spec__.origin),"remotecaller.js"),"r",encoding="utf-8").read()
class Context:
    _clientSocket:Optional[SocketType]
    def __init__(self)->None:
        self._clientSocket=None
    def __enter__(self):
        self.open()
        return self
    def __setitem__(self,key:str,value:Any)->None:
        self.set(key,value)
    def __getitem__(self,key:str)->Any:
        return self.get(key)
    def __exit__(self,exc_type,exc_val,exc_tb):
        self.close()
        return False
    def __del__(self)->None:
        tempSocket=self._clientSocket
        if tempSocket is not None:
            try:
                tempSocket.sendall(b"{}\n")
            except OSError:
                pass
            try:
                tempSocket.close()
            except OSError:
                pass
    def open(self,auto:bool=False)->None:
        if self._clientSocket is not None:
            raise AttributeError("The context has already been opened")
        with socket(AF_INET,SOCK_STREAM) as serverSocket:
            serverPort=bindAvailablePort(serverSocket,1)
            if auto:
                script="\"auto\";\n"+REMOTE_CALLER%(CONFIG["script_host_name"],serverPort)
            else:
                script=REMOTE_CALLER%(CONFIG["script_host_name"],serverPort)
            runString(script,"%s-%d"%(CONFIG["remote_caller_title"],time_ns()))
            self._clientSocket=serverSocket.accept()[0]
    def set(self,key:str,value:Any)->None:
        if self._clientSocket is None:
            raise AttributeError("The context hasn't been opened yet")
        self._clientSocket.sendall((dumps({"key":str(key),"value":value},ensure_ascii=False,separators=(",",":"))+"\n").encode("utf-8"))
    def call(self,function:str,**kwargs:Any)->Any:
        if self._clientSocket is None:
            raise AttributeError("The context hasn't been opened yet")
        keys=[]
        values=[]
        for i,j in kwargs.items():
            keys.append(str(i))
            values.append(j)
        self._clientSocket.sendall((dumps({"args":values,"keys":keys,"target":str(function)},ensure_ascii=False,separators=(",",":"))+"\n").encode("utf-8"))
        with self._clientSocket.makefile("r",encoding="utf-8") as socketReader:
            result=loads(socketReader.readline())
        if "result" in result:
            return result["result"]
        elif "error" in result:
            raise ValueError("The JavaScript function threw an error: %s"%(result["error"],))
    def get(self,key:str)->Any:
        if self._clientSocket is None:
            raise AttributeError("The context hasn't been opened yet")
        self._clientSocket.sendall((dumps({"key":str(key)},ensure_ascii=False,separators=(",",":"))+"\n").encode("utf-8"))
        with self._clientSocket.makefile("r",encoding="utf-8") as socketReader:
            result=loads(socketReader.readline())
        if "result" in result:
            return result["result"]
        elif "error" in result:
            raise KeyError("During the query, an error was thrown: %s"%(result["error"],))
    def close(self)->None:
        tempSocket=self._clientSocket
        if tempSocket is None:
            raise AttributeError("The context hasn't been opened yet")
        try:
            tempSocket.sendall(b"{}\n")
        except OSError:
            pass
        try:
            tempSocket.close()
        except OSError:
            pass
        self._clientSocket=None


#-*-coding:utf-8;-*-
from array import ArrayType,array
from copy import deepcopy
from json import dumps,loads
from os.path import dirname,join
from socket import AF_INET,SOCK_STREAM,SocketType,socket
from threading import Lock,Thread
from time import time_ns
from typing import Any,Callable,Dict,List,Optional,Tuple,Union
from warnings import warn
from .runner import CONFIG,bindAvailablePort,runString
MODULE_PATH=dirname(__spec__.origin)
LOCATOR_SCRIPT=open(join(MODULE_PATH,"locator.js"),"r",encoding="utf-8").read()
RECORDER_SCRIPT=open(join(MODULE_PATH,"recorder.js"),"r",encoding="utf-8").read()
SENSOR_SCRIPT=open(join(MODULE_PATH,"sensor.js"),"r",encoding="utf-8").read()
def locatorAndSensorMain(readLock:Lock,callbackLock:Lock,endCallbackLock:Lock,result:Dict[str,Union[Dict[str,Any],int]],callback:List[Callable[[Dict[str,Any]],None]],endCallback:List[Callable[[],None]],tempSocket:SocketType,arguments:Dict[str,Union[float,int,str]])->None:
    tempCallback=[]
    with tempSocket as clientSocket:
        clientSocket.sendall((dumps(arguments,ensure_ascii=False,separators=(",",":"))+"\n").encode("utf-8"))
        with clientSocket.makefile("r",encoding="utf-8") as socketReader:
            while True:
                try:
                    inputLine=socketReader.readline()
                except OSError:
                    break
                else:
                    if inputLine.endswith("\n"):
                        inputDict=loads(inputLine)
                        with readLock:
                            result["data"]=inputDict
                            if "serial_number" in result:
                                result["serial_number"]+=1
                            else:
                                result["serial_number"]=0
                        with callbackLock:
                            tempCallback.extend(callback)
                        for i in tempCallback:
                            tempDict=deepcopy(inputDict)
                            try:
                                i(tempDict)
                            except Exception as error:
                                warn("A callback function raised a %s: %s"%(type(error).__name__,str(error)))
                        tempCallback.clear()
                    else:
                        break
    with readLock:
        result.clear()
    with endCallbackLock:
        tempCallback.extend(endCallback)
    for i in tempCallback:
        try:
            i()
        except Exception as error:
            warn("An end callback function raised a %s: %s"%(type(error).__name__,str(error)))
def recorderMain(readLock:Lock,callbackLock:Lock,endCallbackLock:Lock,result:Dict[str,Union[bytes,int]],callback:List[Callable[[ArrayType],None]],endCallback:List[Callable[[],None]],tempSocket:SocketType,receiveSize:int,itemSize:int,typeCode:str,arguments:Dict[str,Union[float,int,str]])->None:
    tempCallback=[]
    with tempSocket as clientSocket:
        clientSocket.sendall((dumps(arguments,ensure_ascii=False,separators=(",",":"))+"\n").encode("utf-8"))
        lastLength=0
        with memoryview(bytearray(receiveSize+itemSize-1)) as inputBuffer:
            while True:
                tempBuffer=inputBuffer[lastLength:receiveSize+lastLength]
                try:
                    inputLength=clientSocket.recv_into(tempBuffer,receiveSize)
                except OSError:
                    break
                else:
                    if inputLength>0:
                        totalLength=inputLength+lastLength
                        lastLength=totalLength%itemSize
                        usedLength=totalLength-lastLength
                        if usedLength>0:
                            inputBytes=inputBuffer[:usedLength].tobytes()
                            with readLock:
                                result["data"]=inputBytes
                                if "serial_number" in result:
                                    result["serial_number"]+=1
                                else:
                                    result["serial_number"]=0
                            with callbackLock:
                                tempCallback.extend(callback)
                            for i in tempCallback:
                                inputArray=array(typeCode,inputBytes)
                                try:
                                    i(inputArray)
                                except Exception as error:
                                    warn("A callback function raised a %s: %s"%(type(error).__name__,str(error)))
                            tempCallback.clear()
                            inputBuffer[:lastLength]=inputBuffer[usedLength:totalLength]
                    else:
                        break
    with readLock:
        result.clear()
    with endCallbackLock:
        tempCallback.extend(endCallback)
    for i in tempCallback:
        try:
            i()
        except Exception as error:
            warn("An end callback function raised a %s: %s"%(type(error).__name__,str(error)))
class LocatorRecorderOrSensor:
    _mainLock:Lock
    _readLock:Lock
    _callbackLock:Lock
    _endCallbackLock:Lock
    _result:Dict[str,Union[Dict[str,Any],bytes,int]]
    _callback:List[Callable[[Union[ArrayType,Dict[str,Any]]],None]]
    _endCallback:List[Callable[[],None]]
    _clientSocket:Optional[SocketType]
    def __init__(self)->None:
        self._mainLock=Lock()
        self._readLock=Lock()
        self._callbackLock=Lock()
        self._endCallbackLock=Lock()
        self._result={}
        self._callback=[]
        self._endCallback=[]
        self._clientSocket=None
    def __del__(self)->None:
        tempSocket=self._clientSocket
        if tempSocket is not None:
            try:
                tempSocket.sendall(b"{}\n")
            except OSError:
                pass
    def addCallback(self,callback:Callable[[Union[ArrayType,Dict[str,Any]]],None])->Callable[[Union[ArrayType,Dict[str,Any]]],None]:
        with self._callbackLock:
            self._callback.append(callback)
        return callback
    def clearCallbacks(self)->None:
        with self._callbackLock:
            self._callback.clear()
    def addEndCallback(self,endCallback:Callable[[],None])->Callable[[],None]:
        with self._endCallbackLock:
            self._endCallback.append(endCallback)
        return endCallback
    def clearEndCallbacks(self)->None:
        with self._endCallbackLock:
            self._endCallback.clear()
    def stop(self)->None:
        with self._mainLock:
            tempSocket=self._clientSocket
            if tempSocket is None:
                raise AttributeError("The locator, recorder or sensor hasn't been started yet")
            try:
                tempSocket.sendall(b"{}\n")
            except OSError:
                pass
            self._clientSocket=None
class Location(LocatorRecorderOrSensor):
    @staticmethod
    def requestPermission()->None:
        runString("runtime.requestPermissions([\"access_fine_location\"]);","%s-%d"%(CONFIG["location_permission_title"],time_ns()))
    def start(self,locationProvider:str=CONFIG["default_location_provider"],updateDelay:int=CONFIG["default_locating_delay"])->None:
        usedProvider=str(locationProvider)
        for i in usedProvider:
            if i not in CONFIG["location_provider_characters"]:
                raise ValueError("Invalid location provider name")
        usedDelay=int(updateDelay)
        if usedDelay<0 or usedDelay>CONFIG["max_locating_delay"]:
            raise ValueError("The delay out of range")
        with self._mainLock:
            if self._clientSocket is not None:
                raise AttributeError("The locator has already been started")
            with socket(AF_INET,SOCK_STREAM) as serverSocket:
                serverPort=bindAvailablePort(serverSocket,1)
                runString(LOCATOR_SCRIPT%(CONFIG["script_host_name"],serverPort),"%s-%d"%(CONFIG["locator_title"],time_ns()))
                clientSocket=serverSocket.accept()[0]
            Thread(target=locatorAndSensorMain,args=(self._readLock,self._callbackLock,self._endCallbackLock,self._result,self._callback,self._endCallback,clientSocket,{"delay":usedDelay,"distance":CONFIG["min_locating_distance"],"provider":usedProvider})).start()
            self._clientSocket=clientSocket
    def read(self)->Tuple[int,Dict[str,Any]]:
        with self._readLock:
            if "serial_number" in self._result:
                serialNumber=self._result["serial_number"]
            else:
                serialNumber=-1
            if "data" in self._result:
                result=deepcopy(self._result["data"])
            else:
                result={}
        return serialNumber,result
class Recorder(LocatorRecorderOrSensor):
    @staticmethod
    def requestPermission()->None:
        runString("runtime.requestPermissions([\"record_audio\"]);","%s-%d"%(CONFIG["record_permission_title"],time_ns()))
    def start(self,audioSource:str=CONFIG["default_record_source"])->None:
        usedSource=str(audioSource)
        for i in usedSource:
            if i not in CONFIG["record_source_characters"]:
                raise ValueError("Invalid audio source name")
        with self._mainLock:
            if self._clientSocket is not None:
                raise AttributeError("The recorder has already been started")
            with socket(AF_INET,SOCK_STREAM) as serverSocket:
                serverPort=bindAvailablePort(serverSocket,1)
                runString(RECORDER_SCRIPT%(CONFIG["script_host_name"],serverPort),"%s-%d"%(CONFIG["recorder_title"],time_ns()))
                clientSocket=serverSocket.accept()[0]
            Thread(target=recorderMain,args=(self._readLock,self._callbackLock,self._endCallbackLock,self._result,self._callback,self._endCallback,clientSocket,CONFIG["max_receive_size"],array(CONFIG["audio_array_type"]).itemsize,CONFIG["audio_array_type"],{"bufferratio":CONFIG["record_buffer_ratio"],"channel":CONFIG["record_channel"],"format":CONFIG["record_format"],"samplerate":CONFIG["record_sample_rate"],"source":usedSource})).start()
            self._clientSocket=clientSocket
    def read(self)->Tuple[int,ArrayType]:
        with self._readLock:
            if "serial_number" in self._result:
                serialNumber=self._result["serial_number"]
            else:
                serialNumber=-1
            if "data" in self._result:
                result=array(CONFIG["audio_array_type"],self._result["data"])
            else:
                result=array(CONFIG["audio_array_type"])
        return serialNumber,result
class Sensor(LocatorRecorderOrSensor):
    def start(self,sensorType:str,sensorDelay:int=CONFIG["default_sensor_delay"])->None:
        usedType=str(sensorType)
        for i in usedType:
            if i not in CONFIG["sensor_type_characters"]:
                raise ValueError("Invalid sensor type name")
        usedDelay=int(sensorDelay)
        if usedDelay<0 or usedDelay>CONFIG["max_sensor_delay"]:
            raise ValueError("The delay out of range")
        with self._mainLock:
            if self._clientSocket is not None:
                raise AttributeError("The sensor has already been started")
            with socket(AF_INET,SOCK_STREAM) as serverSocket:
                serverPort=bindAvailablePort(serverSocket,1)
                runString(SENSOR_SCRIPT%(CONFIG["script_host_name"],serverPort),"%s-%d"%(CONFIG["sensor_title"],time_ns()))
                clientSocket=serverSocket.accept()[0]
            Thread(target=locatorAndSensorMain,args=(self._readLock,self._callbackLock,self._endCallbackLock,self._result,self._callback,self._endCallback,clientSocket,{"delay":usedDelay,"latency":CONFIG["max_sensor_latency"],"type":usedType})).start()
            self._clientSocket=clientSocket
    def read(self)->Tuple[int,Dict[str,Any]]:
        with self._readLock:
            if "serial_number" in self._result:
                serialNumber=self._result["serial_number"]
            else:
                serialNumber=-1
            if "data" in self._result:
                result=deepcopy(self._result["data"])
            else:
                result={}
        return serialNumber,result


#-*-coding:utf-8;-*-
if __name__=="__main__":
    from argparse import ArgumentParser
    from importlib import import_module,reload
    from os.path import abspath,dirname,join
    class AutoFilePath(str):
        pass
    class FilePath(str):
        pass
    configFilePath=join(dirname(__spec__.origin),"config.json")
    parser=ArgumentParser(prog="python3 -m %s"%(__spec__.parent,),description="Launch Auto.js and AutoX.js scripts through Python in Termux.",epilog="You can view the config file at the following path to learn about the config items:\n%s"%(abspath(configFilePath),))
    parser.add_argument("-v","--version",action="version",version="AutoxJS 1.0.14")
    parser.add_argument("-a","--auto",action="append",nargs="+",type=AutoFilePath,help="run Auto.js and AutoX.js auto files",metavar="<file>",dest="script")
    parser.add_argument("-c","--config",action="append",nargs=2,help="modify config values, without changing their data types",metavar=("<key>","<value>"))
    parser.add_argument("-f","--file",action="append",nargs="+",type=FilePath,help="run Auto.js and AutoX.js script files",metavar="<file>",dest="script")
    parser.add_argument("-s","--string",action="append",nargs="+",help="run Auto.js and AutoX.js commands as strings",metavar="<command>",dest="script")
    parseResult=parser.parse_args()
    if parseResult.config:
        from copy import deepcopy
        from json import dump
        runner=import_module(".runner",__spec__.parent)
        configFile=deepcopy(runner.CONFIG)
        for i in parseResult.config:
            configFile[i[0]]=type(configFile[i[0]])(i[1])
            print("Config item \"%s\" has been changed to \"%s\""%(i[0],i[1]))
        dump(configFile,open(configFilePath,"w",encoding="utf-8"),ensure_ascii=False,indent=2)
        if parseResult.script:
            runner=reload(runner)
            for i in parseResult.script:
                for j in i:
                    if isinstance(j,AutoFilePath):
                        runner.runAutoFile(str(j))
                    elif isinstance(j,FilePath):
                        runner.runFile(str(j))
                    else:
                        runner.runString(j)
    elif parseResult.script:
        runner=import_module(".runner",__spec__.parent)
        for i in parseResult.script:
            for j in i:
                if isinstance(j,AutoFilePath):
                    runner.runAutoFile(str(j))
                elif isinstance(j,FilePath):
                    runner.runFile(str(j))
                else:
                    runner.runString(j)
    else:
        parser.error("no valid arguments entered")


#-*-coding:utf-8;-*-
from setuptools import setup
setup(
    name="AutoxJS",
    version="1.0.14",
    description="Launch Auto.js and AutoX.js scripts through Python in Termux.",
    long_description="Run \"python3 -m autojs -h\" to learn how to use this package.<br/><br/>For new versions of AutoX.js, running \"python3 -m autojs -c intent_component org.autojs.autoxjs/.external.open.RunIntentActivity\" to change the package name may be necessary.",
    long_description_content_type="text/markdown",
    author="Enbuging",
    author_email="electricfan@yeah.net",
    url="https://github.com/CannotLoadName/AutoxJS",
    download_url="https://github.com/CannotLoadName/AutoxJS/releases",
    packages=["autojs"],
    license="MIT License",
    keywords=["Auto.js","AutoX.js","Termux","Android","automation"],
    platforms=["Android","Linux"],
    package_data={"autojs":["autorunner.js","filerunner.js","stringrunner.js","locator.js","recorder.js","sensor.js","remotecaller.js","config.json"]},
    zip_safe=True
)